package main

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"launchpad.net/lpad"
	"os"
	"os/exec"
	"path/filepath"
	"regexp"
	"strings"
)

type BranchInfo struct {
	Path     string
	URL      string
	Parent   *BranchInfo
	tempPath string
}

var pushBranchRe = regexp.MustCompile("push branch: ([^ \n]+)")
var parentBranchRe = regexp.MustCompile("parent branch: ([^ \n]+)")

// readBranchInfo reads information from the branch at path.
func readBranchInfo(path string) (*BranchInfo, error) {
	return _readBranchInfo(path, make(map[string]bool))
}

var branchLoop = fmt.Errorf("branch parent links in loop")

func _readBranchInfo(path string, seen map[string]bool) (*BranchInfo, error) {
	logf("Looking up information for branch: %s", path)
	output, err := exec.Command("bzr", "root", path).CombinedOutput()
	if err != nil {
		return nil, bzrError("bzr root", err, output)
	}
	path = strings.TrimSpace(string(output))
	if seen[path] {
		return nil, branchLoop
	}
	seen[path] = true
	output, err = exec.Command("bzr", "info", path).CombinedOutput()
	if err != nil {
		return nil, bzrError("bzr info", err, output)
	}
	info := &BranchInfo{Path: path}
	g := pushBranchRe.FindSubmatch(output)
	if g != nil {
		info.URL = string(g[1])
	}
	g = parentBranchRe.FindSubmatch(output)
	if g != nil {
		loc := string(g[1])
		if hasScheme(loc) {
			info.Parent = &BranchInfo{URL: loc}
		} else {
			// Don't break on errors here. It's fine to
			// have a broken parent link.
			info.Parent, err = _readBranchInfo(loc, seen)
			if err == nil && info.Parent.URL == "" && info.Parent.Parent != nil && info.Parent.Parent.URL != "" {
				info.Parent.URL = info.Parent.Parent.URL
				info.Parent.Parent = nil
			}
		}
	}
	return info, nil
}

func hasScheme(loc string) bool {
	for _, c := range loc {
		if c == ':' {
			return true
		}
		if c == '/' {
			break
		}
	}
	return false
}

// Push sends changes from info.Path to the info.URL location.
func (info *BranchInfo) Push() error {
	if info.URL == "" {
		return fmt.Errorf("No push location known for branch: %s", info.Path)
	}
	logf("Pushing branch: %s", info.URL)
	output, err := exec.Command("bzr", "push", info.URL).CombinedOutput()
	if err != nil {
		return bzrError("bzr push", err, output)
	}
	logf("Branch pushed successfully.")
	return nil
}

// Nick returns the branch nick.
func (info *BranchInfo) Nick() (string, error) {
	if info.Path == "" {
		return "", fmt.Errorf("Nick: branch is not local: %s", info.URL)
	}
	cmd := fmt.Sprintf("cd %s; bzr nick", info.Path)
	output, err := exec.Command("/bin/sh", "-c", cmd).CombinedOutput()
	if err != nil {
		return "", bzrError("bzr nick", err, output)
	}
	return strings.TrimSpace(string(output)), nil
}

// MustBeClean returns an error if 'bzr status' is not clean.
func (info *BranchInfo) MustBeClean() error {
	if info.Path == "" {
		return fmt.Errorf("Can't check status of remote branch: %s", info.URL)
	}
	logf("Checking if 'bzr status' is clean for %s", info.Path)
	output, err := exec.Command("bzr", "status", info.Path).CombinedOutput()
	if err != nil {
		return bzrError("bzr status", err, output)
	}
	if bytes.Count(output, []byte{'\n'}) == 1 && bytes.Index(output, []byte(`See "bzr shelve --list" for details.`)) >= 0 {
		return nil // Shelves are fine.
	}
	if len(output) > 0 {
		return fmt.Errorf("Branch is not clean:\n-----\n%s-----", output)
	}
	return nil
}

// Merge merges changes from the source branch onto info.
func (info *BranchInfo) Merge(source *BranchInfo) error {
	if info.Path == "" {
		return fmt.Errorf("Merge: branch is not local: %s", info.URL)
	}
	origin := source.Path
	if origin == "" {
		origin = source.URL
	}
	logf("Merging branch %s onto %s", origin, info.Path)
	cmd := fmt.Sprintf("cd %s; bzr merge %s", info.Path, origin)
	output, err := exec.Command("/bin/sh", "-c", cmd).CombinedOutput()
	if err != nil {
		return bzrError("bzr merge", err, output)
	}
	logf("Branch merged successfully.")
	return nil
}

// Commit commits any changes made to the branch using message as log.
func (info *BranchInfo) Commit(message, author string) error {
	if info.Path == "" {
		return fmt.Errorf("Commit: branch is not local: %s", info.Path)
	}
	logf("Committing branch: %s", info.Path)

	f, err := ioutil.TempFile("", "commit-")
	if err == nil {
		_, err = f.Write([]byte(message))
	}
	f.Close()
	defer os.Remove(f.Name())
	if err != nil {
		return fmt.Errorf("Commit: %v", err)
	}
	cmd := exec.Command("bzr", "commit", "-F", f.Name())
	if author != "" {
		cmd.Args = append(cmd.Args, "--author", author)
	}
	cmd.Args = append(cmd.Args, info.Path)
	output, err := cmd.CombinedOutput()
	if err != nil {
		return bzrError("bzr commit", err, output)
	}
	logf("Branch committed successfully.")
	return nil
}

// FetchRemote downloads the branch at info.URL onto a local directory.
// The localNeighbor branch is only used to figure a temporary local
// location on disk that would take benefit of a potential repository
// setup.
// If checkout is true, the branch will be retrieved and stored in a bound
// way, so that commits are directly delivered to the remote site.
func (info *BranchInfo) FetchRemote(localNeighbor *BranchInfo, checkout bool) (*BranchInfo, error) {
	if info.URL == "" {
		return nil, fmt.Errorf("No URL known for branch: %s", info.Path)
	}
	if localNeighbor.Path == "" {
		return nil, fmt.Errorf("FetchRemote: neighbor branch without local path: %s", localNeighbor.URL)
	}
	abspath, err := filepath.Abs(localNeighbor.Path)
	if err != nil {
		return nil, fmt.Errorf("FetchRemote: %v", err)
	}
	base := filepath.Join(abspath, ".bzr", "cobzr")
	if _, err := os.Stat(base); err != nil {
		base, _ = filepath.Split(abspath)
	}
	temp, err := ioutil.TempDir(base, "lbox-")
	if err != nil {
		return nil, fmt.Errorf("FetchRemote: %v", err)
	}
	clone := &BranchInfo{tempPath: temp, URL: info.URL, Path: filepath.Join(temp, "lbox")}
	logf("Fetching remote branch: %s", info.URL)
	var cmd *exec.Cmd
	var errCmd string
	if checkout {
		errCmd = "bzr checkout --lightweight"
		cmd = exec.Command("bzr", "checkout", "--lightweight", info.URL, clone.Path)
	} else {
		errCmd = "bzr branch"
		cmd = exec.Command("bzr", "branch", info.URL, clone.Path)
	}
	output, err := cmd.CombinedOutput()
	if err != nil {
		clone.RemoveFetched()
		return nil, bzrError(errCmd, err, output)
	}
	return clone, nil
}

// RemoveFetched removes the temporary directory created by FetchRemote
// for holding the local branch. Calling this method without without
// FetchRemote having been called will do nothing.
func (info *BranchInfo) RemoveFetched() error {
	if info.tempPath == "" {
		return nil
	}
	logf("Removing temporary directory: %s", info.tempPath)
	err := os.RemoveAll(info.tempPath)
	if err != nil {
		return fmt.Errorf("RemoveFetched: %v", err)
	}
	return nil
}

func loadBranchInfos(root *lpad.Root, sourcePath, targetLoc string, adopt bool) (sourceInfo, targetInfo *BranchInfo, err error) {
	sourceInfo, err = readBranchInfo(sourcePath)
	if err != nil {
		return
	}
	if err = sourceInfo.MustBeClean(); err != nil {
		return
	}
	if adopt {
		if sourceInfo.Parent == nil || sourceInfo.Parent.URL == "" {
			err = fmt.Errorf("Adopting needs branch with parent URL.")
			return
		}
		sourceInfo.URL = sourceInfo.Parent.URL
		sourceInfo.Parent = nil
	} else {
		if targetLoc != "" && (targetLoc[0] == '/' || targetLoc[0] == '.') {
			targetInfo, err = readBranchInfo(targetLoc)
			if err != nil {
				return
			}
		} else if targetLoc != "" {
			targetInfo = &BranchInfo{URL: targetLoc}
		} else if sourceInfo.Parent != nil {
			targetInfo = sourceInfo.Parent
		}
		if sourceInfo.URL == "" {
			sourceInfo.URL, err = inferPushURL(root, sourceInfo, targetInfo)
			if failed(err) {
				return
			}
		}
	}
	return
}

func inferPushURL(root *lpad.Root, sourceInfo, targetInfo *BranchInfo) (url string, err error) {
	logf("Infering branch URL for local branch: %s", sourceInfo.Path)
	sibling := siblingURL(sourceInfo)
	if sibling == "" {
		sibling = siblingURL(targetInfo)
	}
	if sibling == "" {
		return "", fmt.Errorf("Can't infer URL for branch.")
	}
	errs := make(chan error, 3)
	var projectName, ownerName, branchName string
	go func() {
		// Take project name from a sibling branch.
		target, err := loadBranch(root, sibling)
		if err == nil {
			projectName = target.ProjectName()
		}
		errs <- err
	}()
	go func() {
		// Take owner name from Launchpad authentication.
		me, err := root.Me()
		if err == nil {
			ownerName = me.Name()
		}
		errs <- err
	}()
	go func() {
		// Take branch name from its nick.
		branchName, err = sourceInfo.Nick()
		errs <- err
	}()

	err = firstError(3, errs)
	if err != nil {
		return "", err
	}
	url = fmt.Sprintf("lp:~%s/%s/%s", ownerName, projectName, branchName)
	logf("Branch URL inferred: %s", url)
	return url, nil
}

func siblingURL(info *BranchInfo) string {
	for info != nil {
		if info.URL != "" {
			return info.URL
		}
		info = info.Parent
	}
	return ""
}

func bzrError(cmd string, err error, output []byte) error {
	if len(output) > 0 {
		return fmt.Errorf("Failed to run %q: %v\n-----\n%s-----", cmd, err, output)
	}
	return fmt.Errorf("Failed to run %q: %v", cmd, err)
}

func checkBranch(branch *BranchInfo) error {
	script := filepath.Join(branch.Path, ".lbox.check")
	if _, err := os.Stat(script); err != nil {
		if os.IsNotExist(err) {
			logf("No check script at %s", script)
			return nil
		}
		return err
	}
	logf("Running %s", script)
	cmd := exec.Command(script)
	cmd.Dir = branch.Path
	output, err := cmd.CombinedOutput()
	if err != nil {
		return fmt.Errorf("Branch check failed: %v\n-----\n%s-----", err, output)
	}
	logf("Branch check succeeded.", script)
	return nil
}
