package main

import (
	"bufio"
	"flag"
	"fmt"
	"io"
	"launchpad.net/goetveld/rietveld"
	"launchpad.net/lpad"
	"log"
	"os"
	"path/filepath"
	"strings"
	"sync"
)

var globalUsage = `
Usage: lbox <command> [options]

The following commands are available. For more details on the
command run lbox <command> -h.

    propose - Push and propose the current branch for merging
    submit  - Merge and commit the current branch
`

var commonFlags = flag.NewFlagSet("common", flag.ExitOnError)

var (
	verbose = commonFlags.Bool("v", false, "Show what is being done in more detail")
	debug   = commonFlags.Bool("debug", false, "If true requests and responses will be dumped")
)

func main() {
	if len(os.Args) < 2 {
		fmt.Fprint(os.Stderr, globalUsage)
		os.Exit(2)
	}
	switch os.Args[1] {
	case "propose":
		propose()
	case "submit":
		submit()
	default:
		fmt.Fprint(os.Stderr, globalUsage)
		os.Exit(2)
	}
	logf("That's all folks.")
}

func parseSubFlags(subcommand string, flags *flag.FlagSet, usage string) {
	commonFlags.VisitAll(func(flag *flag.Flag) {
		flags.Var(flag.Value, flag.Name, flag.Usage)
	})
	flags.Usage = func() {
		fmt.Println(usage)
		flags.PrintDefaults()
		fmt.Println("")
		os.Exit(2)
	}
	parseDefaultFlags(subcommand, flags)
	flags.Parse(os.Args[2:])
	if *verbose || *debug {
		rietveld.SetLogger(logger)
	}
	if *debug {
		lpad.SetDebug(true)
		rietveld.SetDebug(true)
	}
}

func parseDefaultFlags(subcommand string, flags *flag.FlagSet) {
	dir, err := os.Getwd()
	if err != nil {
		panic(err)
	}
	for {
		path := filepath.Join(dir, ".lbox")
		if _, err := os.Stat(path); err == nil {
			args, err := readDefaultArgs(subcommand, path)
			if err != nil {
				failed(fmt.Errorf("Can't read %s: %v", path, err))
				os.Exit(2)
			}
			flags.Parse(args)
			return
		}
		last := dir
		dir = filepath.Dir(dir)
		if dir == last {
			return
		}
	}
}

func readDefaultArgs(subcommand string, path string) (args []string, err error) {
	f, err := os.Open(path)
	if err != nil {
		return nil, err
	}
	defer f.Close()
	r := bufio.NewReader(f)
	prefix1 := subcommand + " "
	prefix2 := subcommand + "\t"
	for {
		slice, err := r.ReadSlice('\n')
		if err == io.EOF {
			return nil, nil
		}
		if err != nil {
			return nil, err
		}
		line := string(slice)
		if strings.HasPrefix(line, prefix1) || strings.HasPrefix(line, prefix2) {
			args := strings.TrimSpace(line[len(prefix1):])
			logf("Found default arguments for %q: %s", subcommand, args)
			return strings.Fields(args), nil
		}
	}
	panic("unreachable")
}

var logger = log.New(os.Stdout, "", log.LstdFlags)

// logf formats arguments via Sprintf, and displays them when
// in verbose mode.
func logf(format string, args ...interface{}) {
	if *verbose || *debug {
		logger.Printf(format, args...)
	}
}

// infof formats arguments via Sprintf, and displays them even
// when not in verbose mode.
func infof(format string, args ...interface{}) {
	if *verbose || *debug {
		logf(format, args...)
	} else {
		fmt.Printf(format, args...)
		fmt.Println("")
	}
}

// firstError waits for n values out of the errors channel
// and returns the first non-nil one.
func firstError(n int, errors chan error) (err error) {
	for i := 0; i < n; i++ {
		if e := <-errors; e != nil && err == nil {
			err = e
		}
	}
	return
}

var errorsMutex sync.Mutex
var errorsSeen bool

// failed returns whether err is nil and in case it is not nil
// causes future calls to failedAny to return true.
func failed(err error) bool {
	if err == nil {
		return false
	}
	infof("error: %v", err)
	errorsMutex.Lock()
	errorsSeen = true
	errorsMutex.Unlock()
	return true
}

// failedAny returns whether any error has been previously passed
// through the failed function.
func failedAny() bool {
	errorsMutex.Lock()
	seen := errorsSeen
	errorsMutex.Unlock()
	return seen
}
