package main

import (
	"errors"
	"flag"
	"fmt"
	"launchpad.net/goetveld/rietveld"
	"launchpad.net/lpad"
	"sync"
)

var proposeUsage = `
Usage: lbox propose [options]

The lbox propose command will push the Bazaar branch in the current
directory to Launchpad, and will then open an editor for entering the
summary and description of the change (first line/rest), and finally
will create a bug and a merge proposal associated with the branch.

Options:
`

var proposeFlags = flag.NewFlagSet("propose", flag.ExitOnError)

var (
	proposeCr     = proposeFlags.Bool("cr", false, "Upload change to codereview.appspot.com as well")
	proposeBug    = proposeFlags.Int("bug", 0, "Bug number to associate with merge proposal")
	proposeNewBug = proposeFlags.Bool("new-bug", false, "Create and associate bug with same summary/description")
	proposeReq    = proposeFlags.String("req", "", "Pre-requisite branch the proposed work is based on (lp:...)")
	proposeFor    = proposeFlags.String("for", "", "Target branch the proposed work is to be merged on (lp:...)")
	proposeWip    = proposeFlags.Bool("wip", false, "Send code, set to work-in-progress, and do not mail request")
	proposeBp     = proposeFlags.String("bp", "", "Blueprint to link the proposed work with")
	proposeEdit   = proposeFlags.Bool("edit", false, "Edit previous summary/description before proposing")
)

func propose() {
	parseSubFlags("propose", proposeFlags, proposeUsage)

	// --- Login and grab branch information ---

	root, err := loginLaunchpad()
	if failed(err) {
		return
	}

	sourceInfo, targetInfo, err := loadBranchInfos(root, ".", *proposeFor, false)
	if failed(err) {
		return
	}

	err = checkBranch(sourceInfo)
	if failed(err) {
		return
	}

	// --- Initial data loading step ---

	var (
		me          *lpad.Person
		bug         *lpad.Bug
		project     *lpad.Project
		milestone   *lpad.Milestone
		sourceOwner lpad.Member
		source      *lpad.Branch
		target      *lpad.Branch
		prereq      *lpad.Branch
		mp          *lpad.MergeProposal
		mpreq       *lpad.MergeProposal
		bp          *lpad.Blueprint
		wg          sync.WaitGroup
	)

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		me, err = loadMe(root)
		failed(err)
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		//var err error

		// It's fine to not find the branch here. It may be
		// available only after we push it below.
		source, _ = loadBranch(root, sourceInfo.URL)
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		if *proposeReq != "" {
			prereq, err = loadBranch(root, *proposeReq)
			failed(err)
		}
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		//var err error

		if targetInfo != nil && targetInfo.URL != "" {
			target, _ = loadBranch(root, targetInfo.URL)
			// Don't fail here. There are other ways to find it.
		}
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		if *proposeBug != 0 {
			bug, err = loadBug(root, *proposeBug)
			failed(err)
		}
	}()

	wg.Wait()
	if failedAny() {
		return
	}

	// --- Fix the source and target branch locations in edge cases ---

	if source != nil {
		// If it's somebody else's branch. Push it again under own user.
		if source.OwnerName() != me.Name() {
			sourceInfo.URL, err = inferPushURL(root, sourceInfo, targetInfo)
			if failed(err) {
				return
			}
			source, err = loadBranch(root, sourceInfo.URL)
			if failed(err) {
				return
			}
		}
		// In case the branch has been previously proposed, use its target.
		if itarget, imp, err := inferTarget(source); err == nil {
			mp = imp
			target = itarget
			targetInfo = &BranchInfo{URL: target.Id()}
		}
		// The push below is really bad if this happens by mistake.
		if target != nil && target.Id() == source.Id() {
			failed(fmt.Errorf("Target and source branches are the same: %s", target.Id()))
			return
		}
	}

	// --- Push the work done ---

	err = sourceInfo.Push()
	if failed(err) {
		return
	}
	if source == nil {
		// Retry as Launchpad may take a moment to acknowledge the branch existence.
		for i := 0; i < 3; i++ {
			source, err = loadBranch(root, sourceInfo.URL)
			if err == nil {
				break
			}
		}
		if failed(err) {
			return
		}
	}

	// --- Load more data that requires the source around already ---

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		project, milestone, err = loadProject(root, source.ProjectName())
		if failed(err) {
			return
		}
		if target == nil {
			target, err = loadFocusBranch(project)
			if failed(err) {
				return
			}
			targetInfo = &BranchInfo{URL: target.Id()}
			logf("Using project focus as landing target: %s", targetInfo.URL)
		}
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		sourceOwner, err = loadMember(root, source.OwnerName())
		failed(err)
	}()

	wg.Wait()
	if failedAny() {
		return
	}

	infof("Branches: %s => %s", source.Id(), target.Id())

	if mp == nil {
		mp, err = loadProposal(source, target)
		if err != notFound {
			failed(err)
			return
		}
	}
	if mp != nil {
		prereq, err = mp.PreReq()
		if err != nil && err != lpad.ErrNotFound {
			failed(err)
			return
		}
	}
	if prereq != nil {
		infof("Requires: %s", prereq.Id())
	}

	// --- All data collected. Edit change log. ---

	var oldLog, newLog *Log
	if mp != nil {
		infof("Proposal: %s", mp.WebPage())
		oldLog = parseLog([]byte(mp.Description()))
	}
	if oldLog != nil && !*proposeEdit {
		newLog = oldLog
	} else {
		newLog, err = editLog(oldLog)
		if failed(err) {
			return
		}
	}
	for newLog.SummaryTooLong() {
		newLog, err = editLog(newLog)
		if failed(err) {
			return
		}
	}
	if !*proposeCr && newLog.CodeReviewId != 0 {
		// Branch is already linked to a codereview, so use -cr implicitly.
		*proposeCr = true
	}

	// --- Update all Launchpad data and fetch branch for diffing. ---

	gotBug := make(chan bool, 1)

	diffBaseInfo := targetInfo
	if prereq != nil {
		merged, _, err := isMerged(prereq)
		if failed(err) {
			return
		}
		if !merged {
			diffBaseInfo = &BranchInfo{URL: prereq.Id()}
		}
	}

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		if *proposeCr {
			diffBaseInfo, err = diffBaseInfo.FetchRemote(sourceInfo, false)
			failed(err)
		}
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		//var err error

		if prereq != nil {
			// Find MP for prereq.
			mps, err := prereq.LandingTargets()
			if failed(err) {
				return
			}
			found := errors.New("marker")
			err = mps.For(func(mp *lpad.MergeProposal) error {
				b, err := mp.Target()
				if err == nil && b.Id() == target.Id() {
					mpreq = mp
					return found
				}
				return nil
			})
			if err != nil && err != found {
				failed(err)
				return
			}
		}
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		if bug == nil && *proposeNewBug {
			stub := lpad.BugStub{
				Title:       newLog.Summary,
				Description: newLog.Description,
				Target:      project,
			}
			bug, err = createBug(root, &stub)
			failed(err)
		}

		gotBug <- (bug != nil)

		if bug != nil {
			infof("Bug: %s", bug.WebPage())
			wg.Add(2)
			go func() {
				failed(linkBugBranch(bug, source))
				wg.Done()
			}()
			go func() {
				failed(setInProgress(bug, milestone, me))
				wg.Done()
			}()
		}
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		if *proposeBp != "" {
			bp, err = loadBlueprint(root, project, *proposeBp)
			if failed(err) {
				return
			}
			infof("Blueprint: %s", bp.WebPage())

			wg.Add(1)
			go func() {
				defer wg.Done()
				var err error

				err = linkBlueprintBranch(bp, source)
				failed(err)
			}()

			wg.Add(1)
			go func() {
				defer wg.Done()
				var err error

				if <-gotBug {
					err = linkBlueprintBug(bp, bug)
					failed(err)
				}
			}()
		}
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		if mp == nil {
			mp, err = proposeMerge(project, prereq, source, target, newLog, *proposeWip)
			if failed(err) {
				return
			}
			infof("Proposal: %s", mp.WebPage())
		} else {
			wg.Add(2)
			go func() {
				defer wg.Done()
				if newLog.Changed(oldLog) {
					failed(setProposalLog(mp, newLog))
				}
			}()
			go func() {
				defer wg.Done()
				status := lpad.StNeedsReview
				if *proposeWip {
					status = lpad.StWorkInProgress
				}
				failed(setProposalStatus(mp, status))
			}()
		}
	}()

	wg.Wait()
	defer diffBaseInfo.RemoveFetched()
	if failedAny() {
		return
	}

	// --- Deliver the diff to Rietveld and mail a PTAL request. ---

	const ptal = "Please take a look."

	var issue *rietveld.Issue
	if *proposeCr && mp != nil {
		creating := newLog.CodeReviewId == 0
		issue, err = sendCodeReview(sourceInfo, diffBaseInfo, newLog, mp, mpreq)
		if failed(err) {
			return
		}
		infof("Rietveld: %s", codeReviewURL(issue))

		wg.Add(1)
		go func() {
			defer wg.Done()
			var err error

			if !*proposeWip {
				err = addCodeReviewComment(issue, ptal, false)
				failed(err)
			}
		}()

		wg.Add(1)
		go func() {
			defer wg.Done()
			var err error

			if creating {
				// Update MP description with the codereview URL.
				err = setProposalLog(mp, newLog)
				failed(err)
			}
		}()
		wg.Wait()
		if failedAny() {
			return
		}
	} else if mp != nil && !*proposeWip {
		// Not using codereview, so request review through MP.
		err = addProposalComment(mp, ptal)
		failed(err)
	}
}
