package main

import (
	"flag"
	"fmt"
	"launchpad.net/goetveld/rietveld"
	"launchpad.net/lpad"
	"strings"
	"sync"
)

var submitUsage = `
Usage: lbox submit [options]

The lbox submit command will find the merge proposal associated with
the Bazaar branch in the current directory, and attempt to merge and
commit it into the target branch using the merge proposal's
description as a commit message.

Options:
`

var submitFlags = flag.NewFlagSet("submit", flag.ExitOnError)

var (
	submitAdopt = submitFlags.Bool("adopt", false, "Submit someone else's branch")
)

func submit() {
	parseSubFlags("submit", submitFlags, submitUsage)

	// --- Login and grab branch information ---

	root, err := loginLaunchpad()
	if failed(err) {
		return
	}
	sourceInfo, _, err := loadBranchInfos(root, ".", "", *submitAdopt)
	if failed(err) {
		return
	}
	source, err := loadBranch(root, sourceInfo.URL)
	if failed(err) {
		return
	}
	target, mp, err := inferTarget(source)
	if failed(err) {
		return
	}
	targetInfo := &BranchInfo{URL: target.Id()}

	prereq, err := mp.PreReq()
	if err != nil && err != lpad.ErrNotFound {
		failed(err)
		return
	}

	infof("Branches: %s => %s", source.Id(), target.Id())
	if prereq != nil {
		infof("Requires: %s", prereq.Id())
	}
	infof("Proposal: %s", mp.WebPage())

	var mpreq *lpad.MergeProposal
	if prereq != nil {
		// Can only submit if pre-req is already merged.
		var merged bool
		merged, mpreq, err = isMerged(prereq)
		if failed(err) {
			return
		}
		if !merged {
			failed(fmt.Errorf("Pre-requisite %s not yet merged.", prereq.Id()))
			return
		}
	}

	// --- Edit change log ---

	oldLog := parseLog([]byte(mp.Description()))
	newLog, err := editLog(oldLog)
	if failed(err) {
		return
	}
	for newLog.SummaryTooLong() {
		newLog, err = editLog(newLog)
		if failed(err) {
			return
		}
	}
	message := newLog.String()

	// --- Update log and load data from Launchpad and Rietveld ---

	var (
		author string
		issue  *rietveld.Issue
	)

	var wg sync.WaitGroup

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		if newLog.Changed(oldLog) {
			err = setProposalLog(mp, newLog)
			failed(err)
		}
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		targetInfo, err = targetInfo.FetchRemote(sourceInfo, true)
		failed(err)
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		if newLog.CodeReviewId != 0 {
			issue, err = rietveld.CodeReview.Issue(newLog.CodeReviewId)
			if failed(err) {
				return
			}
			infof("Rietveld: %s", codeReviewURL(issue))
			message = fmt.Sprintf("%s\nR=%s\nCC=%s\n%s\n",
				message,
				strings.Join(removeMp(issue.ReviewerNicks), ", "),
				strings.Join(removeMp(issue.CcNicks), ", "),
				rietveld.CodeReview.IssueURL(issue),
			)
		}
	}()

	wg.Add(1)
	go func() {
		defer wg.Done()
		var err error

		sourceOwner, err := source.Owner()
		if err == nil {
			author, err = loadMemberAddress(sourceOwner)
		}
		failed(err)
	}()

	wg.Wait()
	defer targetInfo.RemoveFetched()
	if failedAny() {
		return
	}

	// --- Send the final diff and merge ---

	if newLog.CodeReviewId != 0 && !*submitAdopt {
		_, err = sendCodeReview(sourceInfo, targetInfo, newLog, mp, mpreq)
		if failed(err) {
			return
		}
	}

	err = targetInfo.Merge(sourceInfo)
	if failed(err) {
		return
	}

	err = checkBranch(targetInfo)
	if failed(err) {
		return
	}

	err = targetInfo.Commit(message, author)
	if failed(err) {
		return
	}

	// --- Send note about the change ---

	submitted := "*** Submitted:\n\n" + message
	if newLog.CodeReviewId != 0 {
		err = addCodeReviewComment(issue, submitted, true)
	} else {
		err = addProposalComment(mp, submitted)
	}
	if failed(err) {
		return
	}

	infof("Done.")
}
