package main

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"regexp"
	"strconv"
	"strings"
)

type Log struct {
	Summary      string
	Description  string
	CodeReviewId int
}

func (t *Log) Changed(other *Log) bool {
	if other == nil {
		return true
	}
	return t.Summary != other.Summary || t.Description != other.Description
}

func (t *Log) SummaryTooLong() bool {
	return len(t.Summary) > 50
}

func (t *Log) String() string {
	if t.Description == "" {
		return t.Summary + "\n"
	}
	return fmt.Sprintf("%s\n\n%s\n", t.Summary, t.Description)
}

const (
	templateSummary     = "<enter change summary here>"
	templateDescription = "<enter change description here if sensible>"
	templateTooLong     = "TOO LONG (>50): "
)

func editLog(oldLog *Log) (newLog *Log, err error) {
	logf("Opening editor for change summary and description...")

	editor := os.Getenv("EDITOR")
	if editor == "" {
		editor = "sensible-editor"
	}
	f, err := ioutil.TempFile("", "changelog-")
	if err != nil {
		return nil, fmt.Errorf("Can't create temporary changelog file: %v", err)
	}
	defer os.Remove(f.Name())
	defer f.Close()

	if oldLog == nil {
		_, err = fmt.Fprintf(f, "%s\n\n%s\n", templateSummary, templateDescription)
	} else if oldLog.SummaryTooLong() {
		longLog := *oldLog
		longLog.Summary = templateTooLong + longLog.Summary
		_, err = fmt.Fprintf(f, "%s", &longLog)
	} else {
		_, err = fmt.Fprintf(f, "%s", oldLog)
	}
	if err == nil {
		err = f.Sync()
	}
	if err != nil {
		return nil, fmt.Errorf("Failed to write to temporary file: %v", err)
	}

	cmd := exec.Command("sh", "-c", editor+" "+f.Name())
	cmd.Stdin = os.Stdin
	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr
	err = cmd.Run()
	if err != nil {
		return nil, fmt.Errorf("Can't run editor: %v", err)
	}
	f.Seek(0, 0)
	data, err := ioutil.ReadAll(f)
	newLog = parseLog(data)
	if strings.HasPrefix(newLog.Summary, templateTooLong) {
		newLog.Summary = newLog.Summary[len(templateTooLong):]
	}
	if newLog.Summary == "" || newLog.Summary == templateSummary {
		return nil, fmt.Errorf("Change summary is empty.")
	}
	if newLog.Description == templateDescription {
		newLog.Description = ""
	}
	if oldLog != nil {
		newLog.CodeReviewId = oldLog.CodeReviewId
	}
	infof("-----\n%s-----", newLog)
	return newLog, nil
}

var codeReviewRe = regexp.MustCompile(`\n\nhttps://codereview.appspot.com/(\d+)/[\n ]*$`)

func parseLog(data []byte) *Log {
	text := &Log{}
	if i := bytes.Index(data, []byte{'\n'}); i >= 0 {
		text.Summary = strings.TrimSpace(string(data[:i]))
		data = data[i+1:]
	}
	m := codeReviewRe.FindSubmatchIndex(data)
	if m != nil {
		id, err := strconv.Atoi(string(data[m[2]:m[3]]))
		if err != nil {
			panic(err)
		}
		text.CodeReviewId = id
		data = data[:m[0]]
	}
	text.Description = string(bytes.TrimSpace(data))
	return text
}
